import std/[critbits, enumerate, macros, os, tables]

import bio

import arctic/[load, memory, step, types]

var DefaultBuiltins* {.compileTime.}: CritBitTree[ArcticBuiltin]

macro arctic_builtin(builtin: typed): untyped =
    let nameid = builtin.name
    let namestr = newLit(nameid.strVal)

    result = quote do:
        `builtin`
        DefaultBuiltins[`namestr`] = `nameid`

proc putstrln(state: var ArcticState): ArcticStepResult {.arctic_builtin.} =
    var p = state.stack.pop.p
    while state.memory[p] != 0:
        stdout.write(state.memory[p])
    stdout.write('\n')
    return CONTINUE

proc parseCmdLine(): (File, seq[string]) =
    let argv = commandLineParams()

    if len(argv) > 0:
        return (open(argv[0]), argv[1..^0])
    else:
        return (stdin, @[])



when isMainModule:
    let (ifile, args) = parseCmdLine()
    var state = ifile.readAll.load
    let builtins = DefaultBuiltins

    var argv: seq[int] = @[]
    for arg in args:
        let a = state.memory.allocate(arg.len + 1)
        for i in 0 .. arg.len:
            state.memory[a + i] = arg[i].uint8
        state.memory[a + arg.len] = 0
        argv.add a
    let x = state.memory.allocate((args.len + 1) * 8)
    for (i, a) in enumerate(argv):
        state.memory.write(x + 8 * i, cast[uint64](a.int64).serialize(littleEndian))
    state.memory.write(x + 8 * args.len, 0.uint64.serialize(littleEndian))
    state.stack.add ArcticType(i: 0) # argv
    state.stack.add ArcticType(i: args.len) # argc
    state.stack.add ArcticType(i: 0) # return address: null

    while true:
        case state.step(builtins):
            of CONTINUE:
                continue
            of BREAKPOINT:
                echo state
            of EXIT:
                break
            of ERROR:
                break
